
/*
 * Copyright 2005-2008 NVIDIA Corporation.  All rights reserved.
 */

#include "traverse.h"

TraverseClassData *Traverse::data = NULL;

/*
 * Traverse::Traverse() - Default constructor, used for all objects that are not the class 
 * object
 */

Traverse::Traverse() {
  _classObj = false;
}

/*
 * Traverse::Traverse(bool) - Class object contructor, used for the singleton object that
 * represents the state of the data flow analysis at the beginning/end of the procedure
 * being analysed (for forward/backward analysis)
 */

Traverse::Traverse(bool forward) {
  data = new TraverseClassData;
  data->forward = forward;
  _classObj = true; 
}

/*
 * Traverse::~Traverse() - Generic destructor, handles all cases.
 */

Traverse::~Traverse() {
  if (_classObj) {
    map<INT32, Traverse*>::iterator it;
    map<WN*, Traverse*>::iterator it2;
    for(it = data->targetmap.begin(); it != data->targetmap.end(); it++) {
      delete it->second;
    }
    for(it2 = data->datamap.begin(); it2 != data->datamap.end() ; it2++) {
      delete it2->second;
    }
    delete data;
  }
}

/*
 * lTraverseForwardTree - Tree walker that handles walking for forward data flow analyses
 */

void lTraverseForwardTree(WN* tree, WN *parent, int kidnum, Traverse *& t, bool *pChange, bool apply) {

  while (tree) {
    switch(WN_operator(tree)) {
      case OPR_BLOCK:
        {
          WN* node = WN_first(tree);
          lTraverseForwardTree(node, tree, 0, t, pChange, apply);
        }
      break;
      case OPR_IF:
        {
           lTraverseForwardTree(WN_if_test(tree), tree, 0, t, pChange, apply);
           Traverse *tmp1 = t->Copy();
           Traverse *tmp2 = t->Copy();
           lTraverseForwardTree(WN_then(tree), tree, 1, tmp1, pChange, apply);
           lTraverseForwardTree(WN_else(tree), tree, 2, tmp2, pChange, apply);

           if (WN_then(tree) && WN_else(tree)) {
             tmp1->Merge(tmp2);
             delete t;
             t = tmp1;
           } else if (WN_then(tree)) {
             t->Merge(tmp1);
           } else if (WN_else(tree)) {
             t->Merge(tmp2);
           }

           delete tmp1;
           delete tmp2;
        }
      break;
      case OPR_DO_WHILE:
        {
          Traverse *tmp1 = t->Copy();
          Traverse *tmp2 = NULL;
	  do {
            delete tmp2;
            tmp2 = tmp1->Copy();
            lTraverseForwardTree(WN_while_body(tree), tree, 1, tmp2, pChange, false);
            lTraverseForwardTree(WN_while_test(tree), tree, 0, tmp2, pChange, false);
          } while (tmp1->Merge(tmp2));
          delete tmp2;
          t->Merge(tmp1);
          delete tmp1;
          lTraverseForwardTree(WN_while_body(tree), tree, 1, t, pChange, apply);
          lTraverseForwardTree(WN_while_test(tree), tree, 0, t, pChange, apply);
        }
      break;
      case OPR_WHILE_DO:
        {
          Traverse *tmp1 = t->Copy();
          Traverse *tmp2 = NULL;
          do {
            delete tmp2;
            tmp2 = tmp1->Copy();
            lTraverseForwardTree(WN_while_test(tree), tree, 0, tmp2, pChange, false);
            lTraverseForwardTree(WN_while_body(tree), tree, 1, tmp2, pChange, false);
          } while (tmp1->Merge(tmp2)); 
          delete tmp2;
          t->Merge(tmp1);
          delete tmp1;
          lTraverseForwardTree(WN_while_test(tree), tree, 0, t, pChange, apply);
          if (apply) {
            tmp1 = t->Copy();
            lTraverseForwardTree(WN_while_body(tree), tree, 1, tmp1, pChange, apply);
            delete tmp1;
          }
        }
      break;
    case OPR_DO_LOOP:
        {
          lTraverseForwardTree(WN_start(tree), tree, 1, t, pChange, apply);
          Traverse *tmp1 = t->Copy();
          Traverse *tmp2 = NULL;
          do {
            delete tmp2;
            tmp2 = tmp1->Copy();
            lTraverseForwardTree(WN_end(tree), tree, 2, tmp2, pChange, false);
            lTraverseForwardTree(WN_do_body(tree), tree, 4, tmp2, pChange, false);
            lTraverseForwardTree(WN_step(tree), tree, 3, tmp2, pChange, false);
          } while (tmp1->Merge(tmp2));
          delete tmp2;
          t->Merge(tmp1);
          delete tmp1;
          lTraverseForwardTree(WN_end(tree), tree, 2, t, pChange, apply);
          if (apply) {
            tmp1 = t->Copy();
            lTraverseForwardTree(WN_do_body(tree), tree, 4, tmp1, pChange, apply);
            lTraverseForwardTree(WN_step(tree), tree, 3, tmp1, pChange, apply);
            delete tmp1;
          }
        }
      break;
    case OPR_LABEL:
        {
          Traverse* tomerge = NULL;
          Traverse* labeldata = NULL;
          map<INT32, Traverse*>::iterator it;
          it = Traverse::data->targetmap.find(WN_label_number(tree));
          if (it != Traverse::data->targetmap.end()) {
            tomerge = it->second;
          }          
          map<WN*, Traverse*>::iterator it2;
          it2 = Traverse::data->datamap.find(tree);
          if (it2 != Traverse::data->datamap.end()) {
            labeldata = it2->second;
          }
          if (labeldata == NULL) {
            labeldata = t->Copy();
            Traverse::data->datamap[tree] = labeldata;
            *pChange = true;
          }
          if (tomerge) {
            if (labeldata->Merge(tomerge)) {
              *pChange = true;
            }
          }
          t->Merge(labeldata);
        }
        break;
    case OPR_FALSEBR:
    case OPR_TRUEBR:
        {
          lTraverseForwardTree(WN_kid0(tree), tree, 0, t, pChange, apply);
          Traverse* tmp1 = t->Copy();
          Traverse* tomerge = NULL;
          map<INT32, Traverse*>::iterator it;
          it = Traverse::data->targetmap.find(WN_label_number(tree));
          if (it != Traverse::data->targetmap.end()) {
            tomerge = it->second;
          }
          if (tomerge == NULL) {
            Traverse::data->targetmap[WN_label_number(tree)] = tmp1;
            *pChange = true;
          } else {
            if (tomerge->Merge(tmp1)) {
              *pChange = true;
            }
            delete tmp1;
          }
        }
        break;
    case OPR_GOTO:
        {
          Traverse* tmp1 = t->Copy();
          Traverse* tomerge = NULL;
          map<INT32, Traverse*>::iterator it;
          it = Traverse::data->targetmap.find(WN_label_number(tree));
          if (it != Traverse::data->targetmap.end()) {
            tomerge = it->second;
          }
          if (tomerge == NULL) {
            Traverse::data->targetmap[WN_label_number(tree)] = tmp1;
            *pChange = true;
          } else {
            if (tomerge->Merge(tmp1)) {
              *pChange = true;
            }
            delete tmp1;
          }
          t->MakeEmpty();
        }
        break;
    case OPR_REGION:
    case OPR_REGION_EXIT:
    case OPR_GOTO_OUTER_BLOCK:
    case OPR_SWITCH:
    case OPR_CASEGOTO:
    case OPR_COMPGOTO:
    case OPR_XGOTO:
    case OPR_AGOTO:
    case OPR_ALTENTRY:
        FmtAssert(0,("Control flow statement not yet handled"));
    default:
      {
        for(int ii = 0; ii < WN_kid_count(tree); ii++) {
          lTraverseForwardTree(WN_kid(tree, ii), tree, ii, t, pChange, apply);
        }
      }
    }
    if (apply) {
      if (
          (OPERATOR_is_stmt(WN_operator(tree)) || OPERATOR_is_scf(WN_operator(tree))) &&
          (!(WN_operator(parent) == OPR_FUNC_ENTRY))
         ) {
        WN *prev = WN_prev(tree);
        WN *next = WN_next(tree);
        WN *newnode = t->Apply(tree, parent);
        if (WN_operator(parent) == OPR_BLOCK) {
          if (newnode == NULL) {
            if (prev == NULL) {
              WN_first(parent) = next;
            }
            if (next != NULL) {
              WN_prev(next) = prev;
            }
            if (next == NULL) {
              WN_last(parent) = prev;
            }
            if (prev != NULL) {
              WN_next(prev) = next;            
            }
          } else {
            if (prev == NULL) {
              WN_first(parent) = newnode;
            } else {
              WN_next(prev) = newnode;
            }
            WN_prev(newnode) = prev;
            if (next == NULL) {
              WN_last(parent) = newnode;
            } else {
              WN_prev(next) = newnode;
            }
            WN_next(newnode) = next;
          }
        } else {
          if (newnode == NULL) {
            if (prev == NULL) {
              WN_kid(parent, kidnum) = next;
            }
            if (next != NULL) {
              WN_prev(next) = prev;
            }
            if (prev != NULL) {
              WN_next(prev) = next;
            }
          } else {
            if (prev == NULL) {
              WN_kid(parent, kidnum) = newnode;
            } else {
              WN_next(prev) = newnode;
            }
            WN_prev(newnode) = prev;
            if (next != NULL) {
              WN_prev(next) = newnode;
            }
            WN_next(newnode) = next;
          }
        }
      } else {
        WN_kid(parent, kidnum) = t->Apply(tree, parent);
      }
    }
    t->TransferFunction(tree, parent);

    if (OPERATOR_is_stmt(WN_operator(tree)) || OPERATOR_is_scf(WN_operator(tree))) {
      tree = WN_next(tree);
    } else {
      tree = NULL;
    }
  }
}

/*
 * lTraverseForward() - Entry point for forward analysis, finds the code given a procedure
 */


static void lTraverseForward(WN *tree, Traverse *& t, bool *pChange, bool apply) {

  FmtAssert( WN_operator(tree) == OPR_FUNC_ENTRY, ("Must pass func entry"));

  lTraverseForwardTree(WN_func_body(tree), tree, WN_kid_count(tree) - 1, t, pChange, apply);

}

/*
 * lTraverseBackward() - Entry point for backward analysis, finds the code given a procedure
 */

static WN *lTraverseBackward(WN *tree, Traverse *& t, bool *pChange, bool apply) {

}

/*
 * Traverse::Solve - the main solver to be called on the class object initialized with
 * the state at the beginning/end of the procedure for forward/backward analysis
 */

void Traverse::Solve(WN *tree) {

  // iterate until convergence
  bool change = true;
  Traverse* tmp1 = NULL;
  while (change) {
    change = false;
    delete tmp1;
    tmp1 = this->Copy();
    if (data->forward) {
      lTraverseForward(tree, tmp1, &change, false);
    } else {
      lTraverseBackward(tree, tmp1, &change, false);
    }
  }
  delete tmp1;

  // now apply changes to the tree 
  Traverse* tmp2 = this->Copy();
  if (data->forward) {
    lTraverseForward(tree, tmp2, &change, true);
  } else {
    lTraverseBackward(tree, tmp2, &change, true);
  }
  delete tmp2;

  FmtAssert(change == false, ("cannot have changes during the apply phase"));


}



