
/*
 * Copyright 2005-2008 NVIDIA Corporation.  All rights reserved.
 */

/*
 * This is a generic data flow solver framework to be used with
 * any level of WHIRL nodes, it provides the basic functionality
 * to perform data flow analysis and subsequent transformation 
 * based on the result of analysis on a WHIRL expression tree.
 * The solver requires the user to derive a class from the 
 * traverse class, such class is supposed to contain state that
 * represents the state of the data flow analysis being performed
 * at some point in the WHIRL. Any data structure can be used
 * to store such information. The solver only requires that
 * the following methods are implemented:
 * 
 * Copy() - To make a copy of the current state in the object
 * MakeEmpty() - To make an empty object from the current object
 * Merge() - To merge the data flow information in these objects,
 *           returning true if there was any changes to the state
 * TransferFunction() - To alter the state of the object taking
 *                      into account the current expression
 * Apply() - To apply the transformation of interest using the
 *           dataflow information computed by the solver.
 *
 * The solver assumes the first object of the derived class that
 * is created represents the state of the data flow information
 * at the beginning/end of the procedure being analysed (for
 * forward/backward analysis). This object is also special because
 * this is the object that the method Solve() should be called
 * on to start the process of analysis and transformation.
 */

#ifndef traverse_INCLUDED
#define traverse_INCLUDED

#include "wn.h"
#include <map>

using namespace std;

class Traverse;

struct TraverseClassData {
  bool forward;
  map<INT32, Traverse*> targetmap;
  map<WN*, Traverse*> datamap;
};

class Traverse {
public:
  Traverse();
  Traverse(bool forward);
  virtual ~Traverse();

  virtual Traverse *Copy() = 0;
  virtual void MakeEmpty() = 0;
  virtual bool Merge(Traverse* other) = 0;
  virtual void TransferFunction(WN *node, WN *parent) = 0;
  virtual WN *Apply(WN *node, WN *parent) = 0;

  void Solve(WN *tree);

private:

  static struct TraverseClassData* data;
  bool _classObj;

  friend void lTraverseForwardTree(WN* tree, WN *parent, int kidnum, Traverse *& t, bool *pChange, bool apply);

};

#endif // traverse_INCLUDED
